/*
 * ExportPlanImagePlugin.java
 *
 * Copyright (c) 2021 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.exportplan;

import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Collections;
import java.util.Iterator;
import java.util.Locale;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.concurrent.Callable;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.jnlp.BasicService;
import javax.jnlp.ServiceManager;
import javax.jnlp.UnavailableServiceException;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.KeyStroke;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import com.eteks.sweethome3d.model.Compass;
import com.eteks.sweethome3d.model.Elevatable;
import com.eteks.sweethome3d.model.Home;
import com.eteks.sweethome3d.model.InterruptedRecorderException;
import com.eteks.sweethome3d.model.Level;
import com.eteks.sweethome3d.model.RecorderException;
import com.eteks.sweethome3d.model.Selectable;
import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;
import com.eteks.sweethome3d.swing.AutoCommitSpinner;
import com.eteks.sweethome3d.swing.FileContentManager;
import com.eteks.sweethome3d.swing.PlanComponent;
import com.eteks.sweethome3d.swing.SwingTools;
import com.eteks.sweethome3d.swing.SwingViewFactory;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.viewcontroller.ContentManager;
import com.eteks.sweethome3d.viewcontroller.ContentManager.ContentType;
import com.eteks.sweethome3d.viewcontroller.HomeView;
import com.eteks.sweethome3d.viewcontroller.ThreadedTaskController;
import com.eteks.sweethome3d.viewcontroller.View;

/**
 * A plug-in that generates the image of a home plan.
 * @author Emmanuel Puybaret
 */
public class ExportPlanImagePlugin extends Plugin {
  @Override
  public PluginAction [] getActions() {
    return new PluginAction [] {new ExportPlanImagePluginAction("com.eteks.sweethome3d.plugin.exportplan.ApplicationPlugin",
        "EXPORT_PLAN_TO_IMAGE", getPluginClassLoader(), true)};
  }

  protected class ExportPlanImagePluginAction extends PluginAction {
    private String resourceBaseName;

    public ExportPlanImagePluginAction(String resourceBaseName,  String actionPrefix, ClassLoader pluginClassLoader, boolean enabled) {
      super(resourceBaseName, actionPrefix, pluginClassLoader, enabled);
      this.resourceBaseName = resourceBaseName;
    }

    /**
     * Exports edited home.
     */
    public void execute() {
      final ResourceBundle resource = ResourceBundle.getBundle(this.resourceBaseName,
          Locale.getDefault(), getPluginClassLoader());
      final HomeView homeView = getHomeController().getView();

      try {
        // Ignore plug-in in protected Java Web Start environment
        ServiceManager.lookup("javax.jnlp.FileSaveService");
        // Use an uneditable editor pane to let user select text in dialog
        JEditorPane messagePane = new JEditorPane("text/html",
            resource.getString("exportPlanImageJavaWebStartInfo.message"));
        messagePane.setOpaque(false);
        messagePane.setEditable(false);
        try {
          // Lookup the javax.jnlp.BasicService object
          final BasicService service = (BasicService)ServiceManager.lookup("javax.jnlp.BasicService");
          // If basic service supports  web browser
          if (service.isWebBrowserSupported()) {
            // Add a listener that displays hyperlinks content in browser
            messagePane.addHyperlinkListener(new HyperlinkListener() {
              public void hyperlinkUpdate(HyperlinkEvent ev) {
                if (ev.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                  service.showDocument(ev.getURL());
                }
              }
            });
          }
        } catch (UnavailableServiceException ex) {
          // Too bad : service is unavailable
        }

        String title = resource.getString("exportPlanImageJavaWebStartInfo.title");
        JOptionPane.showMessageDialog((JComponent)homeView, messagePane, title, JOptionPane.WARNING_MESSAGE);
        return;
      } catch (UnavailableServiceException ex) {
      }

      final Home home = getHome().clone();
      home.setSelectedItems(Collections.<Selectable> emptyList());

      final Rectangle2D itemsBounds = new Rectangle2D.Float();
      final int margin = 10;
      final PlanComponent plan = new PlanComponent(home, getUserPreferences(), null) {
          {
            setScale(1);
            // Compute level item bounds with getItemBounds protected method
            Level selectedLevel = home.getSelectedLevel();
            for (Selectable item : home.getSelectableViewableItems()) {
              if (item instanceof Elevatable && ((Elevatable)item).isAtLevel(selectedLevel)
                  || item instanceof Compass) {
                if (itemsBounds.isEmpty()) {
                  itemsBounds.setRect(getItemBounds(null, item));
                } else {
                  itemsBounds.add(getItemBounds(null, item));
                }
              }
            }
          }

          @Override
          public void exportData(OutputStream out, FormatType formatType, Properties settings) throws IOException {
            // Override export to handle image creation
            int width = Integer.parseInt(settings.getProperty("width"));
            int height = Integer.parseInt(settings.getProperty("height"));
            String fileFormat = settings.getProperty("fileFormat");
            boolean jpeg = fileFormat.startsWith("JPEG");
            BufferedImage image = new BufferedImage(width, height,
                jpeg ? BufferedImage.TYPE_INT_RGB : BufferedImage.TYPE_INT_ARGB);
            Graphics2D g2D = (Graphics2D)image.getGraphics();
            if (jpeg) {
              g2D.setColor(getBackgroundColor(PaintMode.EXPORT));
              g2D.fillRect(0, 0, image.getWidth(), image.getHeight());
            }
            g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
            g2D.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
            g2D.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
            g2D.translate(margin, margin);
            float scale = (float)((width - 2 * margin) / itemsBounds.getWidth());
            g2D.scale(scale, scale);
            g2D.translate(-itemsBounds.getMinX(), -itemsBounds.getMinY());
            paintHomeItems(g2D, 1, getBackgroundColor(PaintMode.EXPORT), getForegroundColor(PaintMode.EXPORT), PaintMode.EXPORT);
            g2D.dispose();

            Iterator<ImageWriter> iter = ImageIO.getImageWritersByFormatName(jpeg ? "JPEG" : fileFormat);
            ImageWriter writer = (ImageWriter)iter.next();
            ImageWriteParam writeParam = writer.getDefaultWriteParam();
            if (jpeg) {
              writeParam.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
              writeParam.setCompressionQuality(new Float(fileFormat.substring(fileFormat.lastIndexOf(' ') + 1)));
            }
            writer.setOutput(ImageIO.createImageOutputStream(out));
            writer.write(null, new IIOImage(image, null, null), writeParam);
            writer.dispose();
            out.flush();
          }
        };

      // Create file format label and combo box
      JLabel fileFormatLabel = new JLabel(resource.getString("fileFormatLabel.text"));
      JComboBox<?> fileFormatComboBox = new JComboBox(new Object [] {"PNG", "JPEG 0.3", "JPEG 0.5", "JPEG 0.7", "JPEG 0.9"});
      fileFormatComboBox.setRenderer(new DefaultListCellRenderer() {
          @Override
          public Component getListCellRendererComponent(JList list, Object value,
                                                        int index, boolean isSelected, boolean cellHasFocus) {
            String string = (String)value;
            String displayedValue = "";
            if ("PNG".equals(string)) {
              displayedValue = resource.getString("fileFormatComboBox.png.text");
            } else if (((String)string).startsWith("JPEG")) {
              float compressionQuality = Float.parseFloat(string.substring(string.lastIndexOf(' ') + 1));
              displayedValue = String.format(resource.getString("fileFormatComboBox.jpeg.text"), Math.round(compressionQuality * 100));
            }
            return super.getListCellRendererComponent(list, displayedValue, index, isSelected,
                cellHasFocus);
          }
        });

      // Create width and height labels
      JLabel widthLabel = new JLabel(resource.getString("widthLabel.text"));
      int minWidth;
      int maxWidth;
      int minHeight;
      int maxHeight;
      if (itemsBounds.getWidth() > itemsBounds.getHeight()) {
        minHeight = 2 * margin;
        minWidth = (int)Math.round(minHeight * itemsBounds.getWidth() / itemsBounds.getHeight());
        maxWidth = 10000;
        maxHeight = (int)Math.round(maxWidth * itemsBounds.getHeight() / itemsBounds.getWidth());
      } else {
        minWidth = 2 * margin;
        minHeight = (int)Math.round(minWidth * itemsBounds.getHeight() / itemsBounds.getWidth());
        maxHeight = 10000;
        maxWidth = (int)Math.round(maxHeight * itemsBounds.getWidth() / itemsBounds.getHeight());
      }
      final SpinnerNumberModel widthSpinnerModel = new SpinnerNumberModel(minWidth, minWidth, maxWidth, 10);
      AutoCommitSpinner widthSpinner = new AutoCommitSpinner(widthSpinnerModel);

      JLabel heightLabel = new JLabel(resource.getString("heightLabel.text"));
      final SpinnerNumberModel heightSpinnerModel = new SpinnerNumberModel(minHeight, minHeight, maxHeight, 10);
      AutoCommitSpinner heightSpinner = new AutoCommitSpinner(heightSpinnerModel);
      // Keep width and height proportional to preferred size
      ChangeListener changeListener = new ChangeListener() {
          public void stateChanged(ChangeEvent ev) {
            if (ev.getSource() == widthSpinnerModel) {
              heightSpinnerModel.removeChangeListener(this);
              heightSpinnerModel.setValue(new Integer((int)Math.round(((Number)widthSpinnerModel.getValue()).intValue() * itemsBounds.getHeight() / itemsBounds.getWidth())));
              heightSpinnerModel.addChangeListener(this);
            } else {
              widthSpinnerModel.removeChangeListener(this);
              widthSpinnerModel.setValue(new Integer((int)Math.round(((Number)heightSpinnerModel.getValue()).intValue() * itemsBounds.getWidth() / itemsBounds.getHeight())));
              widthSpinnerModel.addChangeListener(this);
            }
          }
        };
      widthSpinnerModel.addChangeListener(changeListener);
      heightSpinnerModel.addChangeListener(changeListener);
      if (itemsBounds.getWidth() > itemsBounds.getHeight()) {
        widthSpinnerModel.setValue(new Integer(Math.min((int)Math.round(itemsBounds.getWidth()) + 2 * margin, ((Number)widthSpinnerModel.getMaximum()).intValue())));
      } else {
        heightSpinnerModel.setValue(new Integer(Math.min((int)Math.round(itemsBounds.getWidth()) + 2 * margin, ((Number)heightSpinnerModel.getMaximum()).intValue())));
      }

      JPanel exportSettingsPanel = new JPanel(new GridBagLayout());
      int labelAlignment = OperatingSystem.isMacOSX()
          ? JLabel.TRAILING
          : JLabel.LEADING;
      int standardGap = Math.round(5 * SwingTools.getResolutionScale());
      Insets labelInsets = new Insets(0, 0, 0, standardGap);
      // First row
      exportSettingsPanel.add(fileFormatLabel, new GridBagConstraints(
          0, 0, 1, 1, 0, 0, GridBagConstraints.CENTER,
          GridBagConstraints.HORIZONTAL, new Insets(0, 0, standardGap, standardGap), 0, 0));
      fileFormatLabel.setHorizontalAlignment(labelAlignment);
      exportSettingsPanel.add(fileFormatComboBox, new GridBagConstraints(
          1, 0, 3, 1, 0, 0, GridBagConstraints.LINE_START,
          GridBagConstraints.NONE, new Insets(0, 0, standardGap, 0), 0, 0));
      // Second row
      exportSettingsPanel.add(widthLabel, new GridBagConstraints(
          0, 1, 1, 1, 0, 0, GridBagConstraints.CENTER,
          GridBagConstraints.HORIZONTAL, labelInsets, 0, 0));
      widthLabel.setHorizontalAlignment(labelAlignment);
      exportSettingsPanel.add(widthSpinner, new GridBagConstraints(
          1, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START,
          GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 2 * standardGap), 0, 0));
      exportSettingsPanel.add(heightLabel, new GridBagConstraints(
          2, 1, 1, 1, 0, 0, GridBagConstraints.CENTER,
          GridBagConstraints.HORIZONTAL, labelInsets, 0, 0));
      heightLabel.setHorizontalAlignment(labelAlignment);
      exportSettingsPanel.add(heightSpinner, new GridBagConstraints(
          3, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START,
          GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 0), 0, 0));

      if (!OperatingSystem.isMacOSX()) {
        fileFormatLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("fileFormatLabel.mnemonic")).getKeyCode());
        fileFormatLabel.setLabelFor(fileFormatComboBox);
        widthLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("widthLabel.mnemonic")).getKeyCode());
        widthLabel.setLabelFor(widthSpinner);
        heightLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("heightLabel.mnemonic")).getKeyCode());
        heightLabel.setLabelFor(heightSpinner);
      }

      if (SwingTools.showConfirmDialog((JComponent)homeView, exportSettingsPanel,
          resource.getString("exportPlanImageDialog.title"), widthSpinner) == JOptionPane.OK_OPTION) {
        final String fileFormat = (String)fileFormatComboBox.getSelectedItem();
        final int width = ((Number)widthSpinnerModel.getValue()).intValue();
        final int height = ((Number)heightSpinnerModel.getValue()).intValue();

        ContentManager contentManager = new FileContentManager(getUserPreferences()) {
            @Override
            public String showSaveDialog(View        parentView,
                                         String      dialogTitle,
                                         ContentType contentType,
                                         String      path) {
              if (OperatingSystem.isMacOSX()
                  && "Mac App Store".equals(System.getProperty("com.eteks.sweethome3d.deploymentInformation"))) {
                // Use Swing file chooser to avoid crashes
                JFileChooser fileChooser = new JFileChooser();
                fileChooser.setDialogTitle(dialogTitle);
                String selectedFile = null;
                if (fileChooser.showSaveDialog((JComponent)parentView) == JFileChooser.APPROVE_OPTION) {
                  selectedFile = fileChooser.getSelectedFile().getPath();
                  if (!selectedFile.toLowerCase().endsWith(contentType == ContentType.JPEG ? ".jpg" : ".png")) {
                    selectedFile += contentType == ContentType.JPEG ? ".jpg" : ".png";
                  }
                }
                return selectedFile;
              } else {
                return super.showSaveDialog(parentView, dialogTitle, contentType, path);
              }
            }
          };

        // Request a file name
        final String exportedFile = contentManager.showSaveDialog(homeView,
            resource.getString("exportPlanImageDialog.title"), fileFormat.startsWith("JPEG") ? ContentType.JPEG : ContentType.PNG, home.getName());
        if (exportedFile != null) {
          // Export to image in a threaded task
          Callable<Void> exportToHTML5Task = new Callable<Void>() {
              public Void call() throws RecorderException, IOException {
                Properties settings = new Properties();
                settings.put("width", String.valueOf(width));
                settings.put("height", String.valueOf(height));
                settings.put("fileFormat", fileFormat);
                FileOutputStream out = new FileOutputStream(exportedFile);
                plan.exportData(out, null, settings);
                out.close();
                return null;
              }
            };
          ThreadedTaskController.ExceptionHandler exceptionHandler =
              new ThreadedTaskController.ExceptionHandler() {
                public void handleException(Exception ex) {
                  if (!(ex instanceof InterruptedRecorderException)) {
                    ex.printStackTrace();
                    getHomeController().getView().showError(
                        String.format(resource.getString("exportPlanImageError.message"), ex.getMessage()));
                  }
                }
              };
          new ThreadedTaskController(exportToHTML5Task,
              resource.getString("exportPlanImageMessage"), exceptionHandler,
              getUserPreferences(), new SwingViewFactory()).executeTask(homeView);
        }
      }
    }
  }
}
